# -*- coding: utf-8 -*-
"""
aedir - Generic classes and functions for dealing with AE-DIR

Based on python-ldap (see https://www.python-ldap.org

Requires:
- Python 2.7+
- python-ldap 2.4.28+
"""

# Imports from Python's standard lib
import random
import time
import sys
import os
import string
import logging
import logging.handlers
import logging.config
import pprint

# Imports from python-ldap
import ldapurl
import ldap
import ldap.sasl
import ldap.controls
import ldap.modlist
import ldap.resiter
from ldap.dn import escape_dn_chars
from ldap.filter import escape_filter_chars
from ldap.ldapobject import ReconnectLDAPObject
from ldap.syncrepl import SyncreplConsumer
from ldap.resiter import ResultProcessor
from ldap.controls.openldap import SearchNoOpMixIn
from ldap.controls.readentry import PreReadControl, PostReadControl

from aedir.pkginfo import __version__, __author__, __license__

# exported symbols
__all__ = [
    'AEDirUrl',
    'AEDirObject',
    'AEDIR_SEARCH_BASE',
    'aedir_aehost_dn',
    'aedir_aeuser_dn',
    'extract_zone',
    'LoggerFileObj',
    'parse_ldap_config_file',
    'random_string',
    'init_logger',
    'members2uids',
]

# Timeout in seconds when connecting to local and remote LDAP servers
# used for ldap.OPT_NETWORK_TIMEOUT and ldap.OPT_TIMEOUT
LDAP_TIMEOUT = 4.0

# Number of times connecting to LDAP is tried
LDAP_MAXRETRYCOUNT = 2
LDAP_RETRYDELAY = 2.0

# Default search base for AE-DIR
AEDIR_SEARCH_BASE = 'ou=ae-dir'

LDAP_LIMIT_ERRORS = (
    ldap.TIMEOUT,
    ldap.TIMELIMIT_EXCEEDED,
    ldap.SIZELIMIT_EXCEEDED,
    ldap.ADMINLIMIT_EXCEEDED,
)

AESRVGROUP_GROUPREF_ATTRS = [
    'aeDisplayNameGroups',
    'aeSetupGroups',
    'aeLogStoreGroups',
    'aeLoginGroups',
    'aeVisibleGroups',
    'aeVisibleSudoers',
]

# filter template for locating any active entity by name which authenticates
# (e.g. with password)
AUTHC_ENTITY_FILTER_TMPL = (
    '(&'
      '(|'
        '(&'
          '(|(objectClass=aeUser)(objectClass=aeService))'
          '(uid={0})'
        ')'
        '(&'
          '(objectClass=aeHost)'
          '(host={0})'
        ')'
      ')'
      '(aeStatus=0)'
    ')'
)

# default length for generated passwords
PWD_LENGTH = 40

# alphabet used  for generated passwords
PWD_ALPHABET = string.letters + string.digits

# where to find the logging config file
AE_LOGGING_CONFIG = os.environ.get('AE_LOGGING_CONFIG', '/opt/ae-dir/etc/ae-logging.conf')
# which logger to get (qualname)
AE_LOGGER_QUALNAME = os.environ.get('AE_LOGGER_QUALNAME', 'aedir')


def extract_zone(ae_object_dn, aeroot_dn=AEDIR_SEARCH_BASE):
    """
    return the extracted zone name from dn
    """
    asserted_suffix = ','+aeroot_dn.lower()
    if not ae_object_dn.lower().endswith(asserted_suffix):
        raise ValueError(
            '%r does not end with %r' % (ae_object_dn, asserted_suffix)
        )
    return ldap.dn.explode_dn(ae_object_dn[0:-len(aeroot_dn)-1])[-1][3:]


def aedir_aeuser_dn(uid, zone=None, aeroot_dn=AEDIR_SEARCH_BASE):
    """
    Returns a bind DN of a aeUser entry
    """
    if not zone:
        try:
            uid, zone = uid.split('@', 1)
        except ValueError:
            pass
    if zone:
        first_dn_part = ldap.escape_str(
            escape_dn_chars,
            'uid=%s,cn=%s',
            uid,
            zone,
        )
    else:
        first_dn_part = ldap.escape_str(
            escape_dn_chars,
            'uid=%s',
            uid,
        )
    return ','.join((first_dn_part, aeroot_dn))


def aedir_aehost_dn(fqdn, srvgrp=None, zone=None, aeroot_dn=AEDIR_SEARCH_BASE):
    """
    Returns a bind DN of a aeHost entry
    """
    if zone and srvgrp:
        first_dn_part = ldap.escape_str(
            escape_dn_chars,
            'host=%s,cn=%s,cn=%s',
            fqdn,
            srvgrp,
            zone
        )
    else:
        first_dn_part = ldap.escape_str(
            escape_dn_chars,
            'host=%s',
            fqdn
        )
    return ','.join((first_dn_part, aeroot_dn))


def aedir_aegroup_dn(aegroup_cn, aeroot_dn=AEDIR_SEARCH_BASE):
    """
    Returns a bind DN of a aeHost entry
    """
    zone_cn, _ = aegroup_cn.split('-', 1)
    return 'cn={0},cn={1},{2}'.format(
        aegroup_cn,
        zone_cn,
        aeroot_dn,
    )


class AEDirUrl(ldapurl.LDAPUrl):
    """
    LDAPUrl class for AE-DIR with some more LDAP URL extensions
    """
    attr2extype = {
        'who':'bindname',
        'cred':'X-BINDPW',
        'trace_level':'trace',
        'pwd_filename':'x-pwdfilename',
        'sasl_mech':'x-saslmech',
    }

    def __init__(
            self,
            ldapUrl=None,
            urlscheme='ldap',
            hostport='',
            dn='',
            attrs=None,
            scope=None,
            filterstr=None,
            extensions=None,
            who=None,
            cred=None,
            trace_level=None,
            sasl_mech=None,
            pwd_filename=None,
        ):
        ldapurl.LDAPUrl.__init__(
            self,
            ldapUrl=ldapUrl,
            urlscheme=urlscheme,
            hostport=hostport,
            dn=dn,
            attrs=attrs,
            scope=scope,
            filterstr=filterstr,
            extensions=extensions,
            who=who,
            cred=cred,
        )
        self.trace_level = trace_level
        self.sasl_mech = sasl_mech
        self.pwd_filename = pwd_filename


class AEDirObject(
        ReconnectLDAPObject,
        ResultProcessor,
        SyncreplConsumer,
        SearchNoOpMixIn,
    ):
    """
    AE-DIR connection class
    """
    aeroot_filter = '(objectClass=aeRoot)'

    def __init__(
            self,
            ldap_url,
            trace_level=0,
            trace_file=None,
            trace_stack_limit=5,
            retry_max=LDAP_MAXRETRYCOUNT,
            retry_delay=LDAP_RETRYDELAY,
            timeout=LDAP_TIMEOUT,
            who=None,
            cred=None,
            cacert_filename=None,
            client_cert_filename=None,
            client_key_filename=None,
            cache_ttl=0.0,
            sasl_authz_id='',
        ):
        """
        Opens connection, sets some LDAP options and binds according
        to what's provided in `uri'.

        Extensions to parameters passed to ReconnectLDAPObject():
        `ldap_url'
          Can contain the full LDAP URL with authc information and base-DN
        `who'
          Bind-DN to be used (overrules ldap_url)
        `cred'
          Bind password to be used (overrules ldap_url)
        """
        # Parse/store LDAP URL into AEDirUrl instance
        if isinstance(ldap_url, unicode):
            ldap_url = ldap_url.encode('utf-8')
        if isinstance(ldap_url, str):
            self.ldap_url_obj = AEDirUrl(ldap_url)
        elif isinstance(ldap_url, AEDirUrl):
            self.ldap_url_obj = ldap_url
        elif ldap_url is None:
            self.ldap_url_obj = AEDirUrl(
                ldapUrl=ldap.get_option(ldap.OPT_URI).split(' ')[0].strip() or None,
                dn=(ldap.get_option(ldap.OPT_DEFBASE) or '').strip() or None,
            )
        else:
            raise ValueError('Invalid value for ldap_url: %r' % ldap_url)
        who = who or self.ldap_url_obj.who
        cred = cred or self.ldap_url_obj.cred
        self.cache_ttl = cache_ttl
        # Initialize the ldap.ldapobject.ReconnectLDAPObject instance
        ReconnectLDAPObject.__init__(
            self,
            self.ldap_url_obj.initializeUrl(),
            trace_level or int(self.ldap_url_obj.trace_level or '0'),
            trace_file,
            trace_stack_limit,
            retry_max,
            retry_delay,
        )
        # Set TLS options
        if self.ldap_url_obj.urlscheme != 'ldapi':
            self.set_tls_options(
                cacert_filename=cacert_filename,
                client_cert_filename=client_cert_filename,
                client_key_filename=client_key_filename,
            )
        # Switch off automatic referral chasing
        self.set_option(ldap.OPT_REFERRALS, 0)
        # Switch off automatic alias dereferencing
        self.set_option(ldap.OPT_DEREF, ldap.DEREF_NEVER)
        # Set timeout values
        self.set_option(ldap.OPT_NETWORK_TIMEOUT, timeout)
        self.set_option(ldap.OPT_TIMEOUT, timeout)
        # Send StartTLS extended operation if clear-text connection was initialized
        if self.ldap_url_obj.urlscheme == 'ldap':
            self.start_tls_s()
        # Bind with authc information found in ldap_url and key-word arguments
        if (self.ldap_url_obj.urlscheme == 'ldapi' and not who) \
           or \
           (self.ldap_url_obj.sasl_mech and self.ldap_url_obj.sasl_mech.upper() == 'EXTERNAL'):
            self.sasl_external_bind_s(authz_id=sasl_authz_id)
        else:
            self.simple_bind_s(who or '', cred or '')
        self.flush_cache()
        # determine the Æ-DIR search base (aeRoot) and cache it
        self._aeroot_dn = None
        self.find_search_base()
        return # __init__()

    def set_tls_options(
            self,
            cacert_filename=None,
            client_cert_filename=None,
            client_key_filename=None,
        ):
        """
        set TLS options for connection even with checking whether cert/key
        files are readable
        """
        # Force server cert validation
        ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_DEMAND)
        # Set path names of TLS related files
        for tls_option, tls_pathname in (
                (ldap.OPT_X_TLS_CACERTFILE, cacert_filename),
                (ldap.OPT_X_TLS_CERTFILE, client_cert_filename),
                (ldap.OPT_X_TLS_KEYFILE, client_key_filename),
            ):
            try:
                if not tls_pathname:
                    # get global default
                    tls_pathname = ldap.get_option(tls_option)
                if tls_pathname:
                    # Check whether CA cert file can be read
                    open(tls_pathname, 'rb')
                    self._l.set_option(tls_option, tls_pathname)
            except ValueError, value_error:
                if sys.platform != 'darwin' and \
                   str(value_error) != 'ValueError: option error':
                    raise
        # this has to be the last option set to let libldap reinitialize TLS context
        self._l.set_option(ldap.OPT_X_TLS_NEWCTX, 0)
        return # set_tls_options()

    def read_rootdse_s(self, filterstr='(objectClass=*)', attrlist=None):
        """
        convenience wrapper around read_s() for reading rootDSE
        """
        ldap_rootdse = self.read_s(
            '',
            filterstr=filterstr,
            attrlist=attrlist or ['*', '+'],
        )
        return ldap_rootdse  # read_rootdse_s()

    def find_search_base(self):
        """
        Returns the aeRoot entry
        """
        if self._aeroot_dn:
            return self._aeroot_dn
        # read namingContexts from rootDSE
        root_dse = self.read_rootdse_s(
            attrlist=['namingContexts', 'defaultNamingContext', 'aeRoot']
        )
        aeroot_dn = root_dse.get('aeRoot', root_dse.get('defaultNamingContext', [None]))[0]
        if aeroot_dn is not None:
            self._aeroot_dn = aeroot_dn
            return aeroot_dn
        naming_contexts = root_dse.get('namingContexts', [])
        while naming_contexts:
            # last element
            naming_context = naming_contexts.pop(-1)
            result = self.search_ext_s(
                naming_context,
                ldap.SCOPE_SUBTREE,
                self.aeroot_filter,
                attrlist=['1.1'],
                sizelimit=2,
            )
            if len(result) != 1:
                raise ldap.NO_UNIQUE_ENTRY(
                    'No or non-unique search result for %r' % (self.aeroot_filter)
                )
            else:
                self._aeroot_dn = result[0][0]
                return result[0][0]
        raise ldap.NO_UNIQUE_ENTRY(
            'aeRoot not found in %r' % (naming_contexts)
        )

    def find_uid(self, uid, attrlist=None):
        """
        Returns a unique aeUser or aeService entry found by uid
        """
        filterstr = ldap.escape_str(
            escape_filter_chars,
            '(uid=%s)',
            uid
        )
        return self.find_unique_entry(
            self._aeroot_dn,
            filterstr=filterstr,
            attrlist=attrlist,
        )

    def find_aegroup(self, common_name, attrlist=None):
        """
        Returns a unique aeGroup entry
        """
        filterstr = ldap.escape_str(
            escape_filter_chars,
            '(&(objectClass=aeGroup)(cn=%s))',
            common_name
        )
        return self.find_unique_entry(
            self._aeroot_dn,
            filterstr=filterstr,
            attrlist=attrlist,
        )

    def find_aehost(self, host_name, attrlist=None):
        """
        Returns a unique aeHost entry found by cn or host
        """
        filterstr = ldap.escape_str(
            escape_filter_chars,
            '(&(objectClass=aeHost)(host=%s))',
            host_name
        )
        return self.find_unique_entry(
            self._aeroot_dn,
            filterstr=filterstr,
            attrlist=attrlist,
        )

    def get_zoneadmins(self, ae_object_dn, attrlist=None, suppl_filter=''):
        """
        Returns LDAP search results of active aeUser entries of all
        zone-admins responsible for the given `ae_object_dn'.
        """
        zone_name = extract_zone(ae_object_dn, aeroot_dn=self._aeroot_dn)
        zone_entry = self.read_s(
            'cn={0},{1}'.format(zone_name, self._aeroot_dn),
            filterstr='(objectClass=aeZone)',
            attrlist=['aeZoneAdmins'],
        )
        if not (zone_entry and 'aeZoneAdmins' in zone_entry):
            return []
        member_of_filter = ''.join([
            '(memberOf={0})'.format(escape_filter_chars(group_dn))
            for group_dn in zone_entry.get('aeZoneAdmins', [])
        ])
        return self.search_s(
            self._aeroot_dn,
            ldap.SCOPE_SUBTREE,
            filterstr=(
                '(&'
                '(objectClass=aeUser)'
                '(aeStatus=0)'
                '(|{0})'
                '{1}'
                ')'
            ).format(
                member_of_filter,
                suppl_filter,
            ),
            attrlist=attrlist,
        ) # get_zoneadmins()

    def get_user_groups(self, uid, memberof_attr=None):
        """
        Gets a set of DNs of aeGroup entries a AE-DIR user (aeUser or
        aeService) is member of
        """
        if memberof_attr:
            attrlist = [memberof_attr]
        else:
            attrlist = ['1.1']
        aeuser_dn, aeuser_entry = self.find_uid(uid, attrlist=attrlist)
        if memberof_attr in aeuser_entry:
            memberof = set([
                v.lower()
                for v in aeuser_entry[memberof_attr]
            ])
        else:
            ldap_result = self.search_s(
                self._aeroot_dn,
                ldap.SCOPE_SUBTREE,
                ldap.escape_str(
                    escape_filter_chars,
                    '(&(objectClass=aeGroup)(member=%s))',
                    aeuser_dn
                ),
                attrlist=['1.1'],
            )
            memberof = set([
                dn.lower()
                for dn, _ in ldap_result
            ])
        return memberof # get_user_groups()

    def get_service_groups(self, service_dn, filterstr='', attrlist=None):
        """
        returns all service group entries the aeHost/aeService defined by
        service_dn is member of
        """
        aeservice_entry = self.read_s(
            service_dn,
            attrlist=['aeSrvGroup']
        ) or {}
        aesrvgroup_dn_list = [','.join(ldap.explode_dn(service_dn)[1:])]
        aesrvgroup_dn_list.extend(aeservice_entry.get('aeSrvGroup', []))
        srvgroup_filter = '(&{0}{1})'.format(
            '(|{0})'.format(''.join([
                '(entryDN={0})'.format(dn)
                for dn in aesrvgroup_dn_list
            ])),
            filterstr,
        )
        return self.search_ext_s(
            self._aeroot_dn,
            ldap.SCOPE_SUBTREE,
            srvgroup_filter,
            attrlist=attrlist or ['1.1'],
        ) or []

    def get_user_srvgroup_relations(
            self,
            uid,
            aesrvgroup_dn,
            ref_attrs=None
        ):
        """
        Gets relationship between a aeUser and a aeSrvGroup object by
        simply returning the matching attributes of the aeSrvGroup entry
        """
        aeuser_memberof = self.get_user_groups(uid)
        ref_attrs = ref_attrs or AESRVGROUP_GROUPREF_ATTRS
        aesrvgroup_entry = self.read_s(
            aesrvgroup_dn,
            filterstr='(objectClass=aeSrvGroup)',
            attrlist=ref_attrs
        )
        if not aesrvgroup_entry:
            raise ValueError('Empty search result for %r' % aesrvgroup_dn)
        srvgroup_relations = []
        for attr_type in ref_attrs:
            if attr_type in aesrvgroup_entry:
                for attr_value in aesrvgroup_entry[attr_type]:
                    if attr_value.lower() in aeuser_memberof:
                        srvgroup_relations.append(attr_type)
                        break
        return srvgroup_relations # get_user_srvgroup_relations()

    def get_user_service_relations(
            self,
            uid,
            service_dn,
            ref_attrs=None
        ):
        """
        get relation(s) between aeUser specified by uid with
        aeHost/aeService specified by service_dn

        returns set of relationship attribute names
        """
        aesrvgroups = self.get_service_groups(service_dn, attrlist=None)
        aesrvgroup_dn_set = set([
            aesrvgroup_dn
            for aesrvgroup_dn, _ in aesrvgroups
        ])
        result = set()
        for aesrvgroup_dn in aesrvgroup_dn_set:
            result.update(
                self.get_user_srvgroup_relations(
                    uid,
                    aesrvgroup_dn=aesrvgroup_dn,
                    ref_attrs=ref_attrs,
                )
            )
        return result

    def search_users(
            self,
            service_dn=None,
            ae_status=0,
            filterstr='(|(objectClass=aeUser)(objectClass=aeService))',
            attrlist=None,
            ref_attr='aeLoginGroups',
            serverctrls=None,
        ):
        """
        starts async search for all aeUser/aeService entries having the
        appropriate relationship on given aeHost/aeService
        """
        aesrvgroups = self.get_service_groups(
            service_dn,
            filterstr='({0}=*)'.format(ref_attr),
            attrlist=[ref_attr],
        )
        ref_attrs_groups = set()
        for _, srvgroup_entry in aesrvgroups:
            ref_attrs_groups.update(map(str.lower, srvgroup_entry[ref_attr]))
        user_group_filter = '(&(aeStatus={0}){1}(|{2}))'.format(
            str(int(ae_status)),
            filterstr,
            ''.join([
                '(memberOf={0})'.format(escape_filter_chars(dn))
                for dn in ref_attrs_groups
            ]),
        )
        msg_id = self.search_ext(
            self._aeroot_dn,
            ldap.SCOPE_SUBTREE,
            user_group_filter,
            attrlist=attrlist,
            serverctrls=serverctrls,
        )
        return msg_id

    def get_users(
            self,
            service_dn,
            ae_status=0,
            filterstr='(|(objectClass=aeUser)(objectClass=aeService))',
            attrlist=None,
            ref_attr='aeLoginGroups',
            serverctrls=None,
        ):
        """
        returns all aeUser/aeService entries having the appropriate
        relationship on given aeHost/aeService  as list of 2-tuples
        """
        msg_id = self.search_users(
            service_dn,
            ae_status=ae_status,
            filterstr=filterstr,
            attrlist=attrlist,
            ref_attr=ref_attr,
            serverctrls=serverctrls,
        )
        return self.result4(msg_id)

    def get_cache_hit_ratio(self):
        """
        Returns percentage of cache hit ratio
        """
        try:
            return (
                100 * float(self._cache_hits) /
                float(self._cache_hits+self._cache_misses)
            )
        except ZeroDivisionError:
            return None

    def search_ext_s(
            self,
            base,
            scope,
            filterstr='(objectClass=*)',
            attrlist=None,
            attrsonly=0,
            serverctrls=None,
            clientctrls=None,
            timeout=-1,
            sizelimit=0,
            cache_ttl=None,
        ):
        """
        caching variant of LDAPObject.search_ext_s()
        """
        if cache_ttl is None:
            cache_ttl = self.cache_ttl
        current_time = time.time()
        cache_key_args = (
            base,
            scope,
            filterstr,
            tuple(attrlist or []),
            attrsonly,
            tuple(serverctrls or []),
            tuple(clientctrls or []),
            timeout,
            sizelimit
        )
        result = None
        if cache_ttl > 0:
            # first look into cache for non-expired search results
            try:
                result_cache_time, result = self._cache[cache_key_args]
            except KeyError:
                pass
            else:
                if current_time > result_cache_time + cache_ttl:
                    # cache entry is expired => remove it no matter what happens later
                    del self._cache[cache_key_args]
                else:
                    self._cache_hits += 1
        if result is None:
            self._cache_misses += 1
            result = ReconnectLDAPObject.search_ext_s(
                self,
                base,
                scope,
                filterstr,
                attrlist,
                attrsonly,
                serverctrls,
                clientctrls,
                timeout,
                sizelimit
            )
            if self.cache_ttl > 0:
                # store result in cache if caching is enabled at all
                self._cache[cache_key_args] = (
                    current_time,
                    result
                )
        return result # search_ext_s()

    def flush_cache(self):
        """
        reset/flush the internal cache
        """
        self._cache_misses = 0
        self._cache_hits = 0
        self._cache = {}

    def uncache_entry(self, entry_dn):
        """
        remove all cached entries related to dn from internal cache
        """
        entry_dn = entry_dn.lower()
        for cache_key_args, cached_results in self._cache.items():
            if cache_key_args[0].lower() == entry_dn:
                del self._cache[cache_key_args]
            else:
                for cached_dn, _ in cached_results[1]:
                    if not cached_dn is None and cached_dn.lower() == entry_dn:
                        del self._cache[cache_key_args]
                        break # stop looping through results

    def get_next_id(self, id_pool_dn=None, id_pool_attr='gidNumber'):
        """
        consumes next ID by sending MOD_INCREMENT modify operation with
        pre-read entry control
        """
        id_pool_dn = id_pool_dn or self._aeroot_dn
        prc = PreReadControl(criticality=True, attrList=[id_pool_attr])
        msg_id = self.modify_ext(
            id_pool_dn,
            [(ldap.MOD_INCREMENT, id_pool_attr, '1')],
            serverctrls=[prc],
        )
        _, _, _, resp_ctrls = self.result3(msg_id)
        return int(resp_ctrls[0].entry[id_pool_attr][0])

    def add_aeuser(
            self,
            zone,
            uid,
            ae_person,
            ae_status=0,
            ae_ticket_id=None,
            description=None,
            aeperson_attrs=None,
            login_shell=None,
            home_directory='/home/{uid}',
            pwd_policy_subentry='cn=ppolicy-users,cn=ae,ou=ae-dir',
        ):
        """
        add a new aeUser entry beneath given zone
        """
        aeperson_attrs = aeperson_attrs or ['mail', 'sn', 'givenName']
        # read the associated aePerson entry
        ae_person_entry = self.read_s(
            ae_person,
            filterstr='(aeStatus=0)',
            attrlist=aeperson_attrs,
        )
        if not ae_person_entry:
            raise ldap.NO_SUCH_OBJECT('no active and readable aePerson entry %r' % ae_person)
        # get new POSIX-UID/GID
        posix_id = str(self.get_next_id())
        # DN of new entry
        add_dn = aedir_aeuser_dn(uid, zone=zone, aeroot_dn=self._aeroot_dn)
        # generate modlist (entry's content)
        add_modlist = [
            ('objectClass', [
                'account',
                'person',
                'organizationalPerson',
                'inetOrgPerson',
                'aeObject',
                'aeUser',
                'posixAccount',
                'ldapPublicKey',
            ]),
            ('aeStatus', [str(ae_status)]),
            ('uid', [uid]),
            ('aePerson', [ae_person]),
            ('uidNumber', [posix_id]),
            ('gidNumber', [posix_id]),
            ('cn', [
                '{givenName} {sn}'.format(
                    givenName=ae_person_entry['givenName'][0],
                    sn=ae_person_entry['sn'][0],
                )
            ]),
            ('homeDirectory', [home_directory.format(uid=uid)]),
            ('displayName', [
                '{givenName} {sn} ({uid}/{uidNumber})'.format(
                    givenName=ae_person_entry['givenName'][0],
                    sn=ae_person_entry['sn'][0],
                    uid=uid,
                    uidNumber=posix_id,
                )
            ]),
            ('pwdPolicySubentry', [pwd_policy_subentry]),
        ]
        for attr_type, attr_values in ae_person_entry.items():
            add_modlist.append((attr_type, attr_values))
        for attr_type, attr_value in (
                ('aeTicketId', ae_ticket_id),
                ('description', description),
                ('loginShell', login_shell),
            ):
            if attr_value:
                add_modlist.append((attr_type, [attr_value]))
        msg_id = self.add_ext(
            add_dn,
            add_modlist,
            serverctrls=[PostReadControl(
                criticality=True,
                attrList=['*', '+']
            )],
        )
        _, _, _, resp_ctrls = self.result3(msg_id)
        return resp_ctrls[0].entry

#    def add_aehost(self, hostname, srvgroup, host_password=None):

#    def modify_aehost_password(self, hostname, host_password):


class LoggerFileObj(object):
    """
    file-like helper class which sends log lines to a logging instance
    """

    def __init__(self, logger, logging_level):
        self._logger = logger
        self._logging_level = logging_level
        self.log_lines = 0

    def write(self, msg):
        """
        write msg to log
        """
        self._logger.log(self._logging_level, msg[:-1])
        self.log_lines += 1

    def read(self, size=None):
        """
        dummy read() method which raises IOError
        """
        raise IOError('File object of class %s is write-only' % (self.__class__.__name__))


def members2uids(members):
    """
    transforms list of group member DNs into list of uid values
    """
    return [
        dn[4:].split(',')[0]
        for dn in members
    ]


def random_string(alphabet=PWD_ALPHABET, length=PWD_LENGTH):
    """
    Returns a random password byte string consisting of `length' bytes of
    chars found in `alphabet'.
    """
    sys_rand = random.SystemRandom()
    random_upper_bound = len(alphabet) - 1
    return ''.join([
        alphabet[sys_rand.randint(0, random_upper_bound)]
        for _ in range(length)
    ])


def parse_ldap_config_file(config_filename, sep=None):
    """
    This function parses a sssd.conf or ldap.conf client configuration file
    and returns a tuple of various parameters.

    Invoke it like this:
    uri_list, who, cred, sasl_mech, cacert_filename = parse_config_file('/path/to/ldap.conf')
    """
    config_file_basename = os.path.basename(config_filename)
    if config_file_basename.startswith('sssd.conf') or \
       config_file_basename.endswith('sssd.conf'):
        guess_sep = '='
        uri_sep = ','
    else:
        guess_sep = ' '
        uri_sep = ' '
    sep = sep or guess_sep
    config_file = open(config_filename, 'rb')
    uri_list = []
    who = None
    cred = None
    sasl_mech = None
    cacert_filename = None
    cert_filename = None
    key_filename = None
    for line in config_file.readlines():
        try:
            key, val = line.strip().split(sep, 1)
        except (IndexError, ValueError):
            key, val = '', ''
        key = key.strip().lower()
        val = val.strip()
        if key in ('ldap_uri', 'uri', 'ldap_backup_uri'):
            # split URI list
            uri_list.extend([
                u.strip() for u in val.split(uri_sep)
            ])
            # filter empty artefacts e.g. caused by double-spaces
            uri_list = [u for u in uri_list if u]
        elif key == 'ldap_default_bind_dn' or key == 'binddn':
            who = val
        elif key == 'ldap_default_authtok' or key == 'bindpw':
            cred = val
        elif key == 'ldap_tls_cacert' or key == 'tls_cacertfile':
            cacert_filename = val
        elif key == 'ldap_tls_cert' or key == 'tls_cert':
            cert_filename = val
        elif key == 'ldap_tls_key' or key == 'tls_key':
            key_filename = val
    return uri_list, who, cred, sasl_mech, cacert_filename, cert_filename, key_filename


def init_logger(
        log_name=None,
        logging_config=AE_LOGGING_CONFIG,
        logger_qualname=AE_LOGGER_QUALNAME,
    ):
    """
    get logger instance
    """
    logging.config.fileConfig(logging_config)
    logger = logging.getLogger(logger_qualname)
    if log_name:
        logger.name = log_name
    return logger
